'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const fs = require('fs');
const path = require('path');

const DOT = '.'.charCodeAt(0);

/*
  Retrieves a list of entries recursively, including file names and stats.
*/
async function listDir (dir, opts = {}) {
  return new Promise((resolve, reject) => {
    _list(dir, opts, (err, records) => {
      if (err) reject(err);
      else resolve(records);
    });
  })
}

function _list (dir, opts, done) {
  let results = [];
  fs.readdir(dir, (err, list) => {
    if (err) return done(err)
    let pending = list.length;
    if (!pending) return done(null, results)
    function _continue () {
      if (!--pending) done(null, results);
    }
    list.forEach((name) => {
      if (opts.ignoreDotFiles && name.charCodeAt(0) === DOT) {
        return _continue()
      }
      let absPath = path.resolve(dir, name);
      fs.stat(absPath, (err, stat) => {
        if (err) return done(err)
        if (stat && stat.isDirectory()) {
          _list(name, opts, (err, res) => {
            if (err) return done(err)
            results = results.concat(res);
            _continue();
          });
        } else {
          results.push(Object.assign({}, stat, {
            name,
            path: absPath
          }));
          _continue();
        }
      });
    });
  });
}

const fs$1 = require('fs');
const path$1 = require('path');

async function isDocumentArchive (archiveDir, opts = {}) {
  // assuming it is a DAR if the folder exists and there is a manifest.xml
  return _fileExists(path$1.join(archiveDir, 'manifest.xml'), opts)
}

function _fileExists (archivePath, opts) {
  return new Promise((resolve, reject) => {
    fs$1.stat(archivePath, (err, stats) => {
      if (err) reject(err);
      else resolve(stats && stats.isFile());
    });
  })
}

const fs$2 = require('fs');

// these extensions are considered to have text content
const TEXTISH = ['txt', 'html', 'xml', 'json'];

/*
  Provides a list of records found in an archive folder.

  @param {object} opts
    - `noBinaryData`: do not load the content of binary files
    - `ignoreDotFiles`: ignore dot-files
    - versioning: set to true if versioning should be enabled
*/
async function readArchive (archiveDir, opts = {}) {
  // make sure that the given path is a dar
  if (await isDocumentArchive(archiveDir, opts)) {
    // first get a list of stats
    const entries = await listDir(archiveDir, opts);
    // then get file records as specified TODO:link
    let resources = {};
    for (var i = 0; i < entries.length; i++) {
      let entry = entries[i];
      let record = await _getFileRecord(entry, opts);
      resources[record.path] = record;
    }
    return {
      resources,
      version: '0'
    }
  } else {
    throw new Error(archiveDir + ' is not a valid document archive.')
  }
}

/*
  Provides a record for a file as it is used for the DocumentArchive persistence protocol.

  Binary files can be exluced using `opts.noBinaryData`.

  @example

  ```
  {
    id: 'manuscript.xml',
    encoding: 'utf8',
    data: '<article>....</article>',
    size: 5782,
    createdAt: 123098123098,
    updatedAt: 123234567890,
  }
  ```
*/
async function _getFileRecord (fileEntry, opts) {
  // for text files load content
  // for binaries use a url
  let record = {
    path: fileEntry.name,
    encoding: null,
    size: fileEntry.size,
    createdAt: fileEntry.birthtime.getTime(),
    updatedAt: fileEntry.mtime.getTime()
  };
  if (_isTextFile(fileEntry.name)) {
    return new Promise((resolve, reject) => {
      fs$2.readFile(fileEntry.path, 'utf8', (err, content) => {
        if (err) return reject(err)
        record.encoding = 'utf8';
        record.data = content;
        resolve(record);
      });
    })
  } else {
    // used internally only
    record._binary = true;
    if (opts.noBinaryContent) {
      return Promise.resolve(record)
    } else {
      return new Promise((resolve, reject) => {
        fs$2.readFile(fileEntry.path, 'hex', (err, content) => {
          if (err) return reject(err)
          record.encoding = 'hex';
          record.data = content;
          resolve(record);
        });
      })
    }
  }
}

function _isTextFile (f) {
  return new RegExp(`\\.(${TEXTISH.join('|')})$`).exec(f)
}

const fs$3 = require('fs');
const path$2 = require('path');

async function writeArchive (archiveDir, rawArchive, opts = {}) {
  let resourceNames = Object.keys(rawArchive.resources);
  let newVersion = '0';

  if (opts.versioning) {
    console.warn('Git based versioning is not yet implemented.');
  }

  return Promise.all(resourceNames.map(f => {
    let record = rawArchive.resources[f];
    let absPath = path$2.join(archiveDir, f);
    switch (record.encoding) {
      case 'utf8': {
        return _writeFile(fs$3, absPath, record.data, 'utf8')
      }
      case 'blob': {
        return _writeFile(fs$3, absPath, record.data)
      }
      // TODO: are there other encodings which we want to support?
      default:
        return false
    }
  })).then(() => {
    return newVersion
  })
}

function _writeFile (fs, p, data, encoding) {
  return new Promise((resolve, reject) => {
    if (typeof data.pipe === 'function') {
      let file = fs.createWriteStream(p);
      data.pipe(file);
      file.on('close', () => {
        resolve();
      });
    } else {
      fs.writeFile(p, data, encoding, (err) => {
        if (err) reject(err);
        else resolve();
      });
    }
  })
}

const fsExtra = require('fs-extra');

async function cloneArchive (archiveDir, newArchiveDir, opts = {}) {
  // make sure that the given path is a dar
  if (await isDocumentArchive(archiveDir, opts)) {
    await fsExtra.copy(archiveDir, newArchiveDir);
    return true
  } else {
    throw new Error(archiveDir + ' is not a valid document archive.')
  }
}

/* global Buffer */

const fs$4 = require('fs');
const path$3 = require('path');

/*
  A storage client optimised for Desktop clients

  NOTE: No versioning is done atm, but users can do a git init in their Dar
  folders.
*/
class FSStorage {
  constructor (rootDir) {
    this._rootDir = rootDir;
  }

  read (archiveDir, cb) {
    archiveDir = this._normalizeArchiveDir(archiveDir);
    readArchive(archiveDir, { noBinaryContent: true, ignoreDotFiles: true })
      .then(rawArchive => {
        // Turn binaries into urls
        Object.keys(rawArchive.resources).forEach(recordPath => {
          let record = rawArchive.resources[recordPath];
          if (record._binary) {
            delete record._binary;
            record.encoding = 'url';
            record.data = path$3.join(archiveDir, record.path);
          }
        });
        cb(null, rawArchive);
      })
      .catch(cb);
  }

  write (archiveDir, rawArchive, cb) {
    archiveDir = this._normalizeArchiveDir(archiveDir);
    _convertBlobs(rawArchive)
      .then(() => {
        return writeArchive(archiveDir, rawArchive)
      })
      .then((version) => {
        cb(null, JSON.stringify({ version }));
      })
      .catch(cb);
  }

  clone (archiveDir, newArchiveDir, cb) {
    archiveDir = this._normalizeArchiveDir(archiveDir);
    newArchiveDir = this._normalizeArchiveDir(newArchiveDir);
    cloneArchive(archiveDir, newArchiveDir)
      .then(success => {
        if (success) cb();
        else cb(new Error('Could not clone archive'));
      })
      .catch(cb);
  }

  _normalizeArchiveDir (archiveDir) {
    if (this._rootDir) {
      archiveDir = path$3.join(this._rootDir, archiveDir);
    }
    return archiveDir
  }
}

/*
  Convert all blobs to array buffers
*/
async function _convertBlobs (rawArchive) {
  let resources = rawArchive.resources;
  let paths = Object.keys(resources);
  for (var i = 0; i < paths.length; i++) {
    let record = resources[paths[i]];
    if (record.encoding === 'blob') {
      record.data = await _blobToArrayBuffer(record.data);
    }
  }
}

function _blobToArrayBuffer (blob) {
  return new Promise((resolve, reject) => {
    // TODO: is there other way to get buffer out of Blob without browser APIs?
    fs$4.readFile(blob.path, (err, buffer) => {
      if (err) return reject(err)
      resolve(buffer);
    });
  })
}

const fs$5 = require('fs');
const fsExtra$1 = require('fs-extra');
const path$4 = require('path');
const yazl = require('yazl');
const yauzl = require('yauzl');

/*
  This storage is used to store working copies of '.dar' files that are located somewhere else on the file-system.
  Texture will first update the working copy, and then updates (rewrites) the `.dar` file.

  The implementation will be done in three major iterations

  Phase I: bare-metal file-system, without versioning etc.
  - open: `dar` file is unpacked into the corresponding internal folder
  - save: internal folder is packed replacing the 'dar' file
  - new: internal folder is created and somehow seeded
  - saveas: internal folder is updated first (like in the current implementation), then cloned into a new internal folder corresponding
      to the new 'dar' file location, and packing the folder into the target 'dar'

  Phase II: basic versioning
  The idea is to have a `.dar` folder within a `dar` file that contains data used to implement versioning. We will use `hyperdrive` for that
  TODO: flesh out the concept

  Phase III: collaboration
  In addition to `hyperdrive` data we will store Texture DAR changes in the `.dar` folder. E.g., this would allow to merge two `dar` files that have a common
  version in their history.

  Status: Phase I
*/
class DarFileStorage {
  constructor (rootDir, baseUrl) {
    this.rootDir = rootDir;
    this.baseUrl = baseUrl;

    this._internalStorage = new FSStorage();
  }

  read (darpath, cb) {
    // console.log('DarFileStorage::read', darpath)
    /*
      - unpack `dar` file as it is into the corresponding folder replacing an existing one
      - only bare-metal fs
    */
    let id = this._path2Id(darpath);
    let wcDir = this._getWorkingCopyPath(id);
    fsExtra$1.removeSync(wcDir);
    fsExtra$1.mkdirpSync(wcDir);
    this._unpack(darpath, wcDir, err => {
      if (err) return cb(err)
      this._internalStorage.read(wcDir, cb);
    });
  }

  write (darpath, rawArchive, cb) { // eslint-disble-line
    let id = this._path2Id(darpath);
    let wcDir = this._getWorkingCopyPath(id);
    this._internalStorage.write(wcDir, rawArchive, err => {
      if (err) return cb(err)
      this._pack(wcDir, darpath, cb);
    });
  }

  clone (darpath, newDarpath, cb) { // eslint-disble-line
    let id = this._path2Id(darpath);
    let wcDir = this._getWorkingCopyPath(id);
    let newId = this._path2Id(newDarpath);
    let newWcDir = this._getWorkingCopyPath(newId);
    this._internalStorage.clone(wcDir, newWcDir, err => {
      if (err) return cb(err)
      this._pack(newWcDir, newDarpath, cb);
    });
  }

  _path2Id (darpath) {
    darpath = String(darpath);
    darpath = path$4.normalize(darpath);
    // convert: '\\' to '/'
    darpath = darpath.replace(/\\+/g, '/');
    // split path into fragments: dir, name, extension
    let { dir, name } = path$4.parse(darpath);
    // ATTENTION: it is probably possible to create collisions here if somebody uses '@' in a bad way.
    // For now, this is acceptable because it is not realistic.
    // Adding an extra slash that got dropped by path.parse().
    dir += '/';
    // replace '/' with '@slash@'
    dir = dir.replace(/\//g, '@slash@');
    // replace ':' with '@colon@'
    dir = dir.replace(/:/g, '@colon@');
    return dir + name
  }

  _getWorkingCopyPath (id) {
    return path$4.join(this.rootDir, id)
  }

  _unpack (darpath, wcDir, cb) {
    // console.log('DarFileStorage::_unpack', darpath, wcDir)
    yauzl.open(darpath, { lazyEntries: true }, (err, zipfile) => {
      if (err) cb(err);
      zipfile.readEntry();
      zipfile.on('entry', (entry) => {
        // dir entry
        if (/\/$/.test(entry.fileName)) {
          zipfile.readEntry();
        // file entry
        } else {
          // console.log('... unpacking', entry.fileName)
          zipfile.openReadStream(entry, (err, readStream) => {
            if (err) throw err
            readStream.on('end', () => {
              zipfile.readEntry();
            });
            let absPath = path$4.join(wcDir, entry.fileName);
            fsExtra$1.ensureDirSync(path$4.dirname(absPath));
            readStream.pipe(fs$5.createWriteStream(absPath));
          });
        }
      });
      zipfile.on('error', err => {
        cb(err);
      });
      zipfile.once('end', () => {
        cb();
      });
    });
  }

  _pack (wcDir, darpath, cb) {
    // console.log('DarFileStorage::_pack')
    let zipfile = new yazl.ZipFile();
    listDir(wcDir).then(entries => {
      for (let entry of entries) {
        let relPath = path$4.relative(wcDir, entry.path);
        // console.log('... adding "%s" as %s', entry.path, relPath)
        zipfile.addFile(entry.path, relPath);
      }
      zipfile.outputStream.pipe(fs$5.createWriteStream(darpath)).on('close', () => {
        cb();
      });
      // call end() after all the files have been added
      zipfile.end();
    }).catch(cb);
  }

  // used by tests
  _getRawArchive (darpath, cb) {
    let id = this._path2Id(darpath);
    let wcDir = this._getWorkingCopyPath(id);
    this._internalStorage.read(wcDir, cb);
  }
}

/**
 * A storage implementation that is bound to a single folder.
 */
class UnpackedDarFolderStorage extends FSStorage {
  constructor (darFolder) {
    super();

    this.darFolder = darFolder;
  }

  _normalizeArchiveDir () {
    return this.darFolder
  }

  clone (archiveDir, newArchiveDir, cb) {
    cb(new Error('Cloning is not supported by this storage type.'));
  }
}

exports.DarFileStorage = DarFileStorage;
exports.FSStorage = FSStorage;
exports.UnpackedDarFolderStorage = UnpackedDarFolderStorage;
exports.cloneArchive = cloneArchive;
exports.readArchive = readArchive;
exports.writeArchive = writeArchive;
//# sourceMappingURL=dar-server.js.map
